/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

obj2map - OBJ file to Quake .map convertor

By Frank Condello
http://www.planetquake.com/paroxysm/
pox@planetquake.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <console.h>
#include <SIOUX.h>
#include "mac_io.h"
#include "mathlib.h"
#include "objlib.h"

#include "obj2map.h"

plane_t			planes[MAX_BRUSHPLANES];
static int		discarded;
	
qboolean		verbose, superverbose, nogroup, useeofs;

FILE 			*out, *details, *ents, **outputfile;
static qboolean	usefloats;
static char		currenttex[256];

static int		currentbrush = 0, numents, currentmode;
static vec3_t	nullvec = { 0, 0, 0 };
static int		faceofs = 4, hmapofs = 0, axiofs = 16, extrofs = 16;
static vec3_t	mdlmaxs, mdlmins, mdlorg;

OBJmodel_t		*model;
vec3_t			translate = {90.0, 0.0, 0.0};// Put Z going up (can overide with command line)

char* ConTypes[] = {"_FAC","_NEX","_HFC","_HEW","_HNS","_AFC","_AEW","_ANS"};

int numConTypes = sizeof(ConTypes)/sizeof(char*);

//
// copyToMap: dump the contents of a temp file to the main map file
//
void copyToMap (char *path)
{
	char		c;
	FILE		*f;
	
	f = Sys_fopen (path, "r");
	
	if (!f)
		Error("Couldn't open %s", path);
	
	do
	{
		c = (char)fgetc(f);
		
		if (c == EOF)
			break;
			
		fprintf (out, "%c", c);
	}
	while (c != EOF);
	
	Sys_fclose (f);
	
	if (!superverbose)// DEBUG - keep temp files...
		Sys_remove (path);
}


//
// setModelBounds: Sets min, maxs and origin for the mesh group passed
//
void setModelBounds (OBJgroup_t	*group)
{
	int 		i, j, k, numverts;
	float 		v[OBJ_MAX_POLYVERTS];
	
	for (i=0 ; i<3 ; i++)
	{
		mdlmins[i] = 999999;
		mdlmaxs[i] = -999999;
	}
	VectorCopy (nullvec, mdlorg);
	
	// Find mins/maxs
	for (i = 0; i < group->numpolys; i++)
	{
		numverts = T(group->polys[i]).numverts;

		for (j = 0; j < numverts; j++)
		{
			for (k = 0; k < 3; k++)
			{
				if (!usefloats)
					v[k] = round(model->vertices[3*T(group->polys[i]).vindices[j]+k]);
				else
					v[k] = model->vertices[3*T(group->polys[i]).vindices[j]+k];
					
				if (v[k] < mdlmins[k]) mdlmins[k] = v[k];
				if (v[k] > mdlmaxs[k]) mdlmaxs[k] = v[k];
			}
		}
	}
	
	// Set the origin
	for (i=0 ; i<3 ; i++)
		mdlorg[i] = (mdlmaxs[i]+mdlmins[i])*0.5;
}


//
// checkEntString: Mesh group names can be used to create spawn points
//
qboolean checkEntString (OBJgroup_t	*group)
{	
	// Player spawn
	if (strstr (group->name, "info_player"))
	{
		setModelBounds(group);
		
		if (!ents)
			ents = Sys_fopen ("ents.tmp", "wa");
			
		fprintf (ents, "\n{\n");
		fprintf (ents, " \"classname\" \"%s\"\n", group->name);
		fprintf (ents, " \"origin\" \"%i %i %i\"\n", (int)round(mdlorg[0]),
													 (int)round(mdlorg[1]),
													 (int)round(mdlorg[2]-8) );
		fprintf (ents, " \"angle\" \"0\"\n");
		fprintf (ents, "}\n");
		
		return true;
	}
	
	return false;
}


//
// startDetailGroup: Start a new detail entity, and set the conversion mode
//
void startDetailGroup (int mode, int contents)
{
	currentmode = mode;
	
	if (nogroup) return;

	if (!details)
		details = Sys_fopen ("details.tmp", "wa");

	fprintf (details, "\n{\n");
	fprintf (details, " \"classname\" \"am_detail\"\n");
	fprintf (details, " \"_contents\" \"%i\"\n", contents);
	
	outputfile = &details;			
}


//
// checkPlane: Don't write redundant planes (most editors/compilers will trash these anyway)
//
qboolean checkPlane (float *v)
{
	static int		numplanes;
	static int		lastbrush;
	int				i, j;
	vec3_t			t1, t2, t3, temp;
	float			d;

   	// Compare existing planes with the current one
   	if (lastbrush == currentbrush)
   	{	
		for (i = 0; i < numplanes; i++)
		{
			for (j = 0 ; j < 3 ; j++)
	   		{
	   			temp[0]=v[j*3+X]; temp[1]=v[j*3+Y]; temp[2]=v[j*3+Z];// dot macro expects a vec3_t
				d = DotProduct(temp, planes[i].normal) - planes[i].dist;
				
				if (d < -ON_EPSILON || d > ON_EPSILON)
					break;
			}
			
			// Already exists, scrap it...	
			if (j == 3) {
				discarded++;
   				return true;
			}
		}
	}
	else // First plane in current brush
	{
		discarded =	numplanes = 0;
		lastbrush = currentbrush;
	}
	
	// Save v as a vector / dist plane
	for (i = 0 ; i < 3 ; i++)
	{
		t1[i] = v[0+i] - v[3+i];
		t2[i] = v[6+i] - v[3+i];
		t3[i] = v[3+i];
	}
	
	CrossProduct(t1,t2, planes[numplanes].normal);

//	if (VectorCompare (planes[numplanes].normal, vec3_origin))
//		Error ("Brush plane with no normal.");

	VectorNormalize (planes[numplanes].normal);
	planes[numplanes].dist = DotProduct (t3, planes[numplanes].normal);
   	
   	numplanes++;	
	
	if (numplanes > MAX_BRUSHPLANES)
		Error ("Exceeded MAX_BRUSHPLANES.");
	
	return false;
}


//
// writeBrushPlane: Writes a single brush plane to the output file
//
void writeBrushPlane (float *v, FILE *f)
{
	int i;
	
	// Rounding to int's is preferred since minute errors in
	// precision can cause brutal imports in Quiver
	if (!usefloats)
	{
		for (i = 0; i < 9; i++)
			v[i] = round(v[i]);
		
		if (checkPlane(v))
			return;
		
		fprintf(f, "  ( %6.0f %6.0f %6.0f ) ( %6.0f %6.0f %6.0f ) ( %6.0f %6.0f %6.0f ) %s 0 0 0 1 1\n",
						v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8], currenttex );
	}
	else
	{	
		if (checkPlane(v))
			return;
		
		fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) %s 0 0 0 1 1\n",
						v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8], currenttex );
	}
}

//
// createExtrudeBrush: Extrudes a polygon along it's vertex normals
//
void createExtrudeBrush (OBJgroup_t	*group, FILE *f, int offset)
{
	int			i, j, k, numverts, idx, numave, totalverts = 0;
	vec3_t		point, point2, back[3];
	float		*norm;
	float		*v, verts[OBJ_MAX_POLYVERTS];
	float		plane[9];
	int			ofs;
	OBJpoly_t 	*poly;
	extrvert_t	*averaged;
	int 		temp = 0;
	
	if (!offset) ofs = extrofs; else ofs = offset;
		
// We have to go through some flaming hoops to get the extrusions to play nice (a vertex has a normal for each face it's in)
// It's best to go over this bit with your eyes closed... :)

	// Add up the verts in this group
	for (i = 0; i < group->numpolys; i++)
		totalverts += T(group->polys[i]).numverts;
		
	// allocate enough memory to hold all possible extruded points
	// this will only fill up if no vertices are shared (i.e. no adjacent polygons - not likely)
	averaged = malloc(sizeof(extrvert_t)*(totalverts + 1));
	memset (averaged, 0, sizeof(extrvert_t)*(totalverts + 1));
	for (i = 0; i < totalverts; i++)
		averaged[i].idx = -1;
	
	numave = 0;
	
	// Extrude each vertex, in each poly, saving shared verticies in the same slot
	for (i = 0; i < group->numpolys; i++)
	{	
		poly = &T(group->polys[i]);
		numverts = poly->numverts;
		
		for (j = 0; j < numverts; j++)
		{
			// Get the normal and vertex indicies
			idx = poly->vindices[j];
			norm = &model->normals[3*poly->nindices[j]];// Don't care about the index, just the actual vector
			
			VectorCopy ((&model->vertices[3*idx]), point);
			
			// Round the vertex ('cause we do it again later when writing the front plane)
			if (!usefloats) {
				point[X] = round(point[X]); point[Y] = round(point[Y]); point[Z] = round(point[Z]);
			}
			
			VectorCopy (point, point2);	
			
			// Extrude the vertex along the normal
			
			VectorMA (point, ofs, norm, point);// We want to re-use the actual extruded point for adjoining polys
			VectorMA (point2, 256+ofs, norm, point2);// point2 is extruded further to counter rounding errors (for side planes)
			
			// Go through the saved list of verticies to see if the current one is shared
			for (k = 0; k < numave; k++)
			{
				// Already exists, add the new point to average
				if (averaged[k].idx == idx) {
					averaged[k].inpolys++;
					VectorAdd(averaged[k].point, point, averaged[k].point);
					VectorAdd(averaged[k].point2, point2, averaged[k].point2);
					break;
				}
			}
			
			// Doesn't exist, save a new vertex
			if (k == numave) {
				averaged[numave].idx = idx;
				averaged[numave].inpolys = 1;
				VectorCopy(point, averaged[numave].point);
				VectorCopy(point2, averaged[numave].point2);
				numave++;
			}
			
			// Won't happen (better not happen!)
			if (numave > totalverts)
				Error("Not enough verticies allocated!");
		}
	}
	
	if (superverbose) {
		printf ("Verticies in group: %i\n", totalverts);
		printf ("Extruded verticies: %i\n", numave);
	}
	
	// Average and quantize the extruded points
	for (i = 0; i < numave; i++) {
		
		float ave;
		
		ave = 1.0/averaged[i].inpolys;

		averaged[i].point[0] *= ave;
		averaged[i].point[1] *= ave;
		averaged[i].point[2] *= ave;
		
		averaged[i].point2[0] *= ave;
		averaged[i].point2[1] *= ave;
		averaged[i].point2[2] *= ave;
		
		if (!usefloats) {
			averaged[i].point[0] = round(averaged[i].point[0]);
			averaged[i].point[1] = round(averaged[i].point[1]);
			averaged[i].point[2] = round(averaged[i].point[2]);
			
			averaged[i].point2[0] = round(averaged[i].point2[0]);
			averaged[i].point2[1] = round(averaged[i].point2[1]);
			averaged[i].point2[2] = round(averaged[i].point2[2]);
		}
				
		if (averaged[i].inpolys	== 1)
			temp++;		
	}

	if (superverbose)
		printf ("Non-shared verticies: %i\n", temp);
			
	// Now we can (finally) save out the brushes...
	for (i = 0; i < group->numpolys; i++)
	{
		poly = &T(group->polys[i]);
		numverts = poly->numverts;
		
		// Put this poly's verticies into v
		for (j = 0; j < numverts; j++)
		{
			for (k = 0; k < 3; k++)
			{
				if (!usefloats)
					verts[k+3*j] = round(model->vertices[3*poly->vindices[j]+k]);
				else
					verts[k+3*j] = model->vertices[3*poly->vindices[j]+k];
			}
		}
		
		v = verts;
		currentbrush++;
		fprintf(f, " {\n");
		
		// Write the face plane
		writeBrushPlane (v, f);
			
		// Write the edge planes
		for (j = 0; j < numverts; j++)
		{				
			// Find the extruded coord for this vert
			for (k = 0; k < numave; k++)
			{
				if (averaged[k].idx == poly->vindices[j]) {
					
					// save the sideplane extruded point
					VectorCopy (averaged[k].point2, point2);
					
					// Save out extruded points for the capping plane
					if (j < 3) VectorCopy (averaged[k].point, back[j]);
						
					break;
				}
			}

			if (k == numave)
				Error ("Extrusion coord not allocated!");
			
			// Loop back to the first vert when we reach the last one
			if ( j == numverts-1 )
			{
				plane[0] = point2[X]; 	plane[1] = point2[Y]; 	plane[2] = point2[Z];
				plane[3] = v[X]; 		plane[4] = v[Y]; 		plane[5] = v[Z];
				plane[6] = v[X+3*j]; 	plane[7] = v[Y+3*j]; 	plane[8] = v[Z+3*j];
				
				writeBrushPlane(plane, f);
			}
			else
			{	
				plane[0] = point2[X]; 	plane[1] = point2[Y]; 	plane[2] = point2[Z];
				plane[3] = v[X+3+3*j]; 	plane[4] = v[Y+3+3*j]; 	plane[5] = v[Z+3+3*j];
				plane[6] = v[X+3*j]; 	plane[7] = v[Y+3*j]; 	plane[8] = v[Z+3*j];
				
				writeBrushPlane(plane, f);
			}
		}		
		
		// write the capping plane (reverse winding)
		plane[0] = back[2][X]; plane[3] = back[1][X]; plane[6] = back[0][X];
		plane[1] = back[2][Y]; plane[4] = back[1][Y]; plane[7] = back[0][Y];
		plane[2] = back[2][Z]; plane[5] = back[1][Z]; plane[8] = back[0][Z];

		writeBrushPlane(plane, f);
		
		fprintf (f, " }\n");
	}
	
	free(averaged);
}


//
// createFaceBrush: Creates a pyramidal brush from a polygon
//
void createFaceBrush (float *v, OBJpoly_t *poly, FILE *f, int offset)
{
	int		i, j, numverts;
	vec3_t	midpoint;
	float	*norm;
	float	plane[9];
	char 	oldtex[256];
	int		ofs;
	
	if (!offset) ofs = faceofs; else ofs = offset;
		
	currentbrush++;
	fprintf(f, " {\n");
	
	// Write the face plane
	writeBrushPlane (v, f);
	
	norm = &model->facetnorms[3 * poly->findex];
	VectorCopy (nullvec, midpoint);
	numverts = poly->numverts;
	
	// Find the midpoint (barycentre, centroid, centre of mass, whatever...)
	for (i = 0; i < numverts; i++)
	{
		for (j = 0; j < 3; j++)
			midpoint[j] += ( (1.0/numverts) * (v[j+3*i]) );
	}

	// Move the midpoint behind the poly along the face normal	
	VectorMA (midpoint, ofs, norm, midpoint);
	
	// Apply an ignore texture to the "insides"
	strcpy (oldtex, currenttex);
	strcpy (currenttex, "ignore");
	
	// Write the edge planes
	for (i = 0; i < numverts; i++)
	{
		// Loop back to the first vert when we reach the last one
		if ( i == numverts-1 )
		{
			plane[0] = midpoint[X]; plane[1] = midpoint[Y]; plane[2] = midpoint[Z];
			plane[3] = v[X]; 		plane[4] = v[Y]; 		plane[5] = v[Z];
			plane[6] = v[X+3*i]; 	plane[7] = v[Y+3*i]; 	plane[8] = v[Z+3*i];
			
			writeBrushPlane(plane, f);
		}
		else
		{	
			plane[0] = midpoint[X]; plane[1] = midpoint[Y]; plane[2] = midpoint[Z];
			plane[3] = v[X+3+3*i]; 	plane[4] = v[Y+3+3*i]; 	plane[5] = v[Z+3+3*i];
			plane[6] = v[X+3*i]; 	plane[7] = v[Y+3*i]; 	plane[8] = v[Z+3*i];
			
			writeBrushPlane(plane, f);
		}
	}
	
	fprintf (f, " }\n");
	
	// Next face is on the outside - restore the original texture
	strcpy (currenttex, oldtex);	
}


//
// createHeightMapBrush: Extrudes a polygon along +/-<axis> and caps it perpendicular to <axis>
//
void createHeightMapBrush (float *v, OBJpoly_t *poly, int axis, FILE *f, int offset, qboolean ae)
{
	int		i, numverts;
	float	plane[9], endcap, sidecap;
	vec3_t	test;
	float	*norm;
	int		ofs;

	if (!useeofs) {
		if (ae) ofs = axiofs; else ofs = hmapofs;
	} else {
		ofs = offset;
	}	
	
	currentbrush++;
	fprintf(f, " {\n");
	
	// Write the face plane
	writeBrushPlane(v, f);
	
	// Extrude away from the face normal along <axis>...
	
	norm = &model->facetnorms[3 * poly->findex];
	VectorCopy (nullvec, test);
	test[axis] = mdlmaxs[axis]+256;
	
	if (DotProduct (norm, test) < 0) {
		if (ae) endcap = -ofs; else endcap = mdlmins[axis]-ofs;
		sidecap = endcap-256;// over compensate for sides to counter rounding errors
	} else {
		if (ae) endcap = ofs; else endcap = mdlmaxs[axis]+ofs;
		sidecap = endcap+256;// over compensate for sides to counter rounding errors
	}
	
	numverts = poly->numverts;
	
	// Write the side planes...
	
	// To keep it simple, start with the real poly, then replace <axis> with the extruded coord.
	// We don't need to distinguish between axial and terrain here, since the capping plane decides that.
	for (i = 0; i < numverts; i++)
	{
		// Loop back to the first vert when we reach the last one
		if ( i == numverts-1 )
		{
			plane[0] = v[X+3*i]; 	plane[1] = v[Y+3*i]; 	plane[2] = v[Z+3*i];	plane[axis] = sidecap;
			plane[3] = v[X]; 		plane[4] = v[Y]; 		plane[5] = v[Z];
			plane[6] = v[X+3*i]; 	plane[7] = v[Y+3*i]; 	plane[8] = v[Z+3*i];
			
			writeBrushPlane(plane, f);
		}
		else
		{	
			plane[0] = v[X+3*i]; 	plane[1] = v[Y+3*i];	plane[2] = v[Z+3*i];	plane[axis] = sidecap;
			plane[3] = v[X+3+3*i]; 	plane[4] = v[Y+3+3*i]; 	plane[5] = v[Z+3+3*i];
			plane[6] = v[X+3*i]; 	plane[7] = v[Y+3*i]; 	plane[8] = v[Z+3*i];
			
			writeBrushPlane(plane, f);
		}
	}
	
	// Write the capping plane
	
	// reverse the winding order of the original poly
	plane[0] = v[6]; plane[1] = v[7]; plane[2] = v[8];
	plane[3] = v[3]; plane[4] = v[4]; plane[5] = v[5];
	plane[6] = v[0]; plane[7] = v[1]; plane[8] = v[2];
	
	// move the endcap along the proper axis (level out for terrains)
	if (!ae) {
		plane[axis] = endcap; plane[axis+3] = endcap; plane[axis+6] = endcap; // Terrain style
	} else {
		plane[axis] = v[axis+6]+endcap; plane[axis+3] = v[axis+3]+endcap; plane[axis+6] = v[axis]+endcap; // Axial extrusion
	}
	writeBrushPlane(plane, f);
	
	fprintf (f, " }\n");	
}


//
// createBrushes: Load & convert the OBJ model
// NOTE: Round input verts to ensure we're quantized to a 1x1x1 grid if !usefloats 
//
void createBrushes (char *inname, char *outname)
{
	int 		i, j, k, numverts, axis;
	char 		in[128];
	float 		v[OBJ_MAX_POLYVERTS];
	OBJgroup_t	*group;
	int			eofs;
	qboolean	ae = false;
	
	strcpy (in, inname);
	
	// Force the proper extension
	StripExtension (in);
	strcat (in, ".obj");
	
	printf ("Input  File Name: %s\n", in);
	printf ("Output File Name: %s\n\n", outname);
	
	// need to average all vertex normals for NEX extrude to work (360)
	model = LoadOBJModel (in, OBJ_NSIDED, 360, translate);
    	
    if (!model || !model->vertices || model->numvertices < 3 || !model->polys)
    	Error ("obj2map: No valid geometry found in OBJ model.\n");
   
    // Start the map
	out = Sys_fopen (outname, "wa");
	fprintf (out, "{\n");
	fprintf (out, " \"classname\" \"worldspawn\"\n");
	fprintf (out, " \"message\" \"Generated by obj2map\"\n");
	fprintf (out, " \"worldtype\" \"0\"\n");
	
	printf("Converting...\n");
	
	group = model->groups;
	
	while(group)
	{
		// See if this is an entity
		if (checkEntString(group))
		{
			numents++;
			group = group->next;
			continue;
		}
		
		// Set the current texture
		strcpy (currenttex, group->texname);

		if (!strcmp (currenttex, "default"))
			strcpy (currenttex, "city2_1");
		
		// reset some extrusion values
		ae = false;
		useeofs = false;
		eofs = 0;

		// Check for a conversion type flag
		for (i=0; i<numConTypes; i++)
			if (strstr(group->name, ConTypes[i])) break;
		
		switch (i)
		{
			// CNVRT_FACE
			case FAC:
				if (verbose) printf("\nUsing per-face mode...");
				startDetailGroup (CNVRT_FACE, 1);// "Solid" Detail brush
				break;
			
			// CNVRT_EXTR
			case NEX:
				if (verbose) printf("\nUsing normal extrusion mode...");
				startDetailGroup (CNVRT_EXTR, -1);// NOT DETAIL (just neatly group it)
				break;
			
			// CNVRT_HMAP
			case HFC:
			case HEW:
			case HNS:
			case AFC:
			case AEW:
			case ANS:
				setModelBounds(group);
 				startDetailGroup (CNVRT_HMAP, -1);// NOT DETAIL (just neatly group it)	
 				switch (i)
				{
					case HFC:
						if (verbose) printf("\nUsing Floor-Ceiling Heightmap mode...");
		 				axis = Z; break;
					case HEW:
						if (verbose) printf("\nUsing East-West Heightmap mode...");
		 				axis = X; break;
					case HNS:
						if (verbose) printf("\nUsing North-South Heightmap mode...");
		 				axis = Y; break;
					case AFC:
						if (verbose) printf("\nUsing Floor-Ceiling Axial Extrusion mode...");
		 				axis = Z; ae = true; break;
					case AEW:
						if (verbose) printf("\nUsing East-West Axial Extrusion mode...");
		 				axis = X; ae = true; break;
					case ANS:
						if (verbose) printf("\nUsing North-South Axial Extrusion mode...");
		 				axis = Y; ae = true; break;
		 		}
		 		break;
		 	
		 	// CNVRT_DIRECT
		 	default:
			 	if (verbose) printf("\nUsing direct mode...");
				currentmode = CNVRT_DIRECT;
				fprintf(out, " {\n");
				currentbrush++;
				outputfile = &out;
				break;
		}
		
		// Digits after name flags override default offsets
		if (currentmode != CNVRT_DIRECT) {
			
			char 	*ptr;
			char	val[5];
			int		count = 0;
			int		sign = 1;
			
			ptr = (strstr(group->name, ConTypes[i]));
			
			if (!ptr) Error ("Invalid keyword pointer.");// won't happen
			
			ptr+=4;// skip the conversion flag
			
			// Allow negative numbers, but only heightmaps can actually work with them
			if (ptr[0]=='-') {ptr++; sign = -1;}
			
			// Get any numbers that follow
			while ( (ptr[count]>='0' && ptr[count]<='9') && count<3 ) {
				val[count] = ptr[count];
				count++;	
			}
			val[count] = '\0';
			
			// If we got something, make sure it's in range for the contype
			if (val[0] && count) {
				
				eofs = atoi(val)*sign;
				
				if (eofs<1 && (i<HFC || i>HNS)) {
					printf("\nWARNING: Illegal offset value of \"%i\" for group \"%s\" (using default).", eofs, group->name);
					eofs = 0;
				} else {
					useeofs = true;// can't just check for eofs == 0, cause it's valid for heightmaps
					if (verbose) printf("\nOffset override: %i", eofs);
				}
			}
		}
		
		if (!currentmode)
			Error ("Couldn't Determine Conversion mode for object: %s", group->name);
		
		if (verbose)
			printf("\nConverting group: %s\nNumpolys: %i\nTexture: %s\n", group->name, group->numpolys, group->texname);
						
		if (!outputfile)
			Error ("Output file not specified!");
					
		// normal extrude deals with the whole group
		if (currentmode == CNVRT_EXTR)
			createExtrudeBrush (group, *outputfile, eofs);
		else
		{	
			// Convert to brushes group by group, one poly at a time
			for (i = 0; i < group->numpolys; i++)
			{
				numverts = T(group->polys[i]).numverts;
				
				// Put this poly's verticies into v
				for (j = 0; j < numverts; j++)
				{
					for (k = 0; k < 3; k++)
					{
						if (!usefloats)
							v[k+3*j] = round(model->vertices[3*T(group->polys[i]).vindices[j]+k]);
						else
							v[k+3*j] = model->vertices[3*T(group->polys[i]).vindices[j]+k];
					}
				}
				
				switch (currentmode)
				{
				case CNVRT_DIRECT:
					writeBrushPlane (v, *outputfile);
					break;
					
				case CNVRT_FACE:
					createFaceBrush (v, &T(group->polys[i]), *outputfile, eofs);
					break;
					
				case CNVRT_HMAP:
					createHeightMapBrush (v, &T(group->polys[i]), axis, *outputfile, eofs, ae);
					break;
					
				default:
					break;
				}	
			}
		}
				
		if (discarded && verbose)
			printf("Redundant polys: %i\n", discarded);
		
		if (currentmode == CNVRT_DIRECT)
			fprintf (out, " }\n");
		else if (!nogroup)// Finish off the detail/group ent
			fprintf (*outputfile, "}\n");
		
		group = group->next;	
	}
	
	// Finish the worldspawn ent
	fprintf (out, "}\n");
	
	// Copy entities into the map file
	
	if (details) {	
		Sys_fclose (details);
		copyToMap ("details.tmp");
	}
	
	if (ents) {	
		Sys_fclose (ents);
		copyToMap ("ents.tmp");
	}
	
	Sys_fclose (out);
	
	printf ("\nConversion Completed!\n");
	
	if (currentbrush == 1)
		printf ("Wrote 1 brush");
	else
		printf ("Wrote %i brushes", currentbrush);
		
	if (numents == 1)
		printf (" and 1 entity.\n");
	else
		printf (" and %i entities.\n", numents);
	
	FreeOBJModel (model);
}


int main (int argc, char **argv)
{
	int 	i;
	char	outfile[128];
	
	SIOUXSettings.userwindowtitle  = "\pobj2map console log";
	SIOUXSettings.columns          = 104;
	SIOUXSettings.initializeTB     = true;
	SIOUXSettings.standalone       = true;
	SIOUXSettings.setupmenus       = true;
	SIOUXSettings.autocloseonquit  = false;
	SIOUXSettings.asktosaveonclose = false;
	SIOUXSettings.showstatusline   = false;
	
	argc = ccommand(&argv);
		
	// Credit where credt is due...
	printf ("-- obj2map -- http://www.planetquake.com/paroxysm/\n");
	printf ("Wavefront OBJ file to Quake MAP convertor written by Frank Condello, 2001.\n\n");
	printf ("OBJ loader adapted from the GLM library (part of GLUT) Written by Nate Robins, 1997.\n");
	printf ("3D Matrix math functions  2001 by Morgan Aldridge.\n");
	printf ("Misc. Library & Math functions from Quake1 Utils  1996-1997 Id Software Inc.\n");
	printf ("Mac malloc by Jack Jansen (adapted from BSD malloc  1983 Regents of the University of California)\n");
	printf ("Inspired by the raw2map & tri2map terrain generators by Seth Galbraith\n");
	printf ("Mac file I.O. courtesy of David Caldwell, Brad Oliver and the rest of the MacGLQuake Crew.\n\n");		
		
	for (i=1; i < argc; i++)
	{
		outfile[0] = '\0';
		
		if (argv[i][0] != '-')
			break;
		else if (!strcmp (argv[i],"-nogroup"))
			nogroup = true;
		else if (!strcmp (argv[i],"-usefloats"))
			usefloats = true;
		else if ( (!strcmp (argv[i],"-verbose")) || (!strcmp (argv[i],"-v")) )
			verbose = true;
		else if (!strcmp (argv[i],"-debug"))
			verbose = superverbose = true;
		else if (!strcmp (argv[i], "-hoffset"))
		{
			hmapofs = atoi (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-foffset"))
		{
			faceofs = atoi (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-noffset"))
		{
			extrofs = atoi (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-aoffset"))
		{
			axiofs = atoi (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-rotX"))
		{
			translate[0] = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-rotY"))
		{
			translate[1] = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-rotZ"))
		{
			translate[2] = atof (argv[i + 1]);
			i++;
		}
		else
			Error ("obj2map: Unknown option '%s'", argv[i]);
	}
	
	if (i != argc - 2 && i != argc - 1)
		Error ("usage: [-nogroup] [-foffset (n)] [-hoffset (n)] [-aoffset (n)] [-eoffset (n)]\n"
			   "       [-usefloats] [-verbose] [-rotX|Y|Z (n)] sourcefile [destfile]");
	
	if (i != argc - 2)
		strcpy (outfile, argv[i]);
	else
		strcpy (outfile, argv[i+1]);
	
	StripExtension (outfile);
	strcat (outfile, ".map");
	
	createBrushes (argv[i], outfile);
	
	return 0;
}
